@@ -1,7 +1,7 @@
-/* Copyright (C) 2006 Brice Arnould.
+/* Copyright (C) 2006-2008 Brice Arnould.
  *  This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
+ * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
  *
  *  This program is distributed in the hope that it will be useful,
@@ -10,8 +10,8 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
+ * along with this program.  If not, see <http://www.gnu.org/licenses/>.
+ *
  */
 #include "linux.h"
 #include <stdlib.h>
@@ -27,10 +27,11 @@
 #include <sys/types.h>		// opendir()
 #include <dirent.h>		// opendir()
 #include <sys/time.h>		// futimes()
-#include <signal.h>		// sigprocmask, sigsetops
 #include "executive.h"
+#include "signals.h"
 
 
+
 /*  Copy the content of file referenced by in_fd to out_fd
  *  Make file sparse if there's more than gap consecutive '\0',
  * and if gap != 0
@@ -43,7 +44,7 @@
 int
 fcopy (int in_fd, int out_fd, size_t gap)
 {
-  assert (in_fd > -1), assert (out_fd > -1), assert (gap >= 0);
+  assert (in_fd > -1), assert (out_fd > -1);
   size_t buffsize = 65535;
   int *buffer;
   /* Prepare files */
@@ -51,6 +52,9 @@ fcopy (int in_fd, int out_fd, size_t gap)
       || -1 == lseek (out_fd, (off_t) 0, SEEK_SET)
       || -1 == ftruncate (out_fd, (off_t) 0))
     return -1;
+  /* Optimisation (on Linux it double the readahead window) */
+  posix_fadvise (in_fd, (off_t) 0, (off_t) 0, POSIX_FADV_SEQUENTIAL);
+  posix_fadvise (in_fd, (off_t) 0, (off_t) 0, POSIX_FADV_WILLNEED);
   /* Get a buffer... */
   {
     if (gap)
@@ -77,7 +81,7 @@ fcopy (int in_fd, int out_fd, size_t gap)
       }
     buffer = alloca (buffsize);	// better than "goto freeall"... or not ?
     // Else would read uninitialised datas when filesize < buffsize
-    // and is_empty wouldn't be set correctly.
+    // and is_empty could be set uncorrectly.
     memset (buffer, 0xFF, buffsize);
   }
   /* Let's go ! */
@@ -120,11 +124,12 @@ fcopy (int in_fd, int out_fd, size_t gap)
 	      empty_buffs++;	// (can't overflow, see cant_wait two lines down)
 	  }
 	/* if in sparse mode, we'll wait for eof or data, or int overflow before writing */
-	cant_wait = !is_empty || eof || (empty_buffs + 1) * buffsize < 0;
+	cant_wait = !is_empty || eof
+	  || (empty_buffs + 1) * buffsize > INT_MAX;
 	if (gap && cant_wait)
 	  {
 	    /* Should we make a hole ? */
-	    if (empty_buffs >= gap)
+	    if (empty_buffs >= gap && len != 0)	// Don't finish with a hole if len == 0
 	      {
 		if (-1 ==
 		    lseek (out_fd, (off_t) (empty_buffs * buffsize),
@@ -168,19 +173,12 @@ fcopy (int in_fd, int out_fd, size_t gap)
 }
 
 
-/* Mark a file as shaked and block signals
+/* Mark a file as shaked
  */
 static void
 capture (struct accused *a, struct law *l)
 {
   assert (a && l);
-  /* Block signals */
-  {
-    sigset_t sset;
-    sigfillset (&sset);
-    sigdelset (&sset, SIGTSTP);
-    sigprocmask (SIG_BLOCK, &sset, NULL);
-  }
   /* Update position time */
   if (l->xattr && -1 == set_ptime (a->fd))
     {
@@ -190,7 +188,7 @@ capture (struct accused *a, struct law *l)
   return;
 }
 
-/* Restore mtime of a file then unblock signals
+/* Restore mtime of a file
  */
 static void
 release (struct accused *a, struct law *l)
@@ -206,12 +204,6 @@ release (struct accused *a, struct law *l)
     tv[1].tv_usec = 0;
     futimes (a->fd, tv);
   }
-  /* unblock signals */
-  {
-    sigset_t sset;
-    sigfillset (&sset);
-    sigprocmask (SIG_UNBLOCK, &sset, NULL);
-  }
   return;
 }
 
@@ -225,16 +217,27 @@ shake_reg (struct accused *a, struct law *l)
   assert (a && l);
   assert (S_ISREG (a->mode)), assert (a->guilty);
   const uint GAP = MAGICLEAP * 4;
+  char *msg;
   if (l->pretend)
     return 0;
   capture (a, l);
-  /* Make a backup */
-  if (-1 == fcopy (a->fd, l->tmpfd, MAGICLEAP))
+  if (-1 ==
+      asprintf (&msg,
+		"%s: unrecoverable internal error ! file has been saved at %s",
+		a->name, l->tmpname))
+    msg = NULL;
+  /* Error handling */
+  if (msg && -1 == fcopy (a->fd, l->tmpfd, MAGICLEAP))
     {
       int errsv = errno;
       unlink (l->tmpname);	// could work
-      error (1, errsv, "%s: temporary copy failed", a->name);
+      if (msg)
+	error (1, errsv, "%s: temporary copy failed", a->name);
+      else
+	error (1, errsv, "%s: failed to initialise failure manager", a->name);
     }
+  /* Disable most signals (except critical ones) */
+  restrict_signals (msg);
   /*  Ask the FS to put the file at a new place, without losing metadatas
    * nor hard links. Works on ReiserFS but should be tested on other filesystems
    */
@@ -246,7 +249,10 @@ shake_reg (struct accused *a, struct law *l)
   if (-1 == fcopy (l->tmpfd, a->fd, GAP))
     error (1, errno, "%s: restore failed ! file have been saved at %s",
 	   a->name, l->tmpname);
+  /* Restore most signals */
+  restore_signals ();
   release (a, l);
+  free (msg);
   return 0;
 }
 
@@ -279,6 +285,7 @@ atimesort (const void *a, const void *b)
 
 /* Return an array containing file names in the named directory,
  * excepted "." and "..".
+ * Return NULL in case of error.
  * If sort is true, file names are sorted by atime.
  * LIMIT : INT_MAX files
  */
@@ -286,13 +293,16 @@ char **
 list_dir (char *name, int sort)
 {
   assert (name);
-  const size_t BUFFSTEP = 32;
+  const size_t BUFFSTEP = 128;
   size_t namel = strlen (name);
   /* Read entries one by one and store their name in a buffer */
   uint n = 0;
   char **buff = malloc (sizeof (*buff) * BUFFSTEP);
   if (!buff)
-    error (1, errno, "%s: malloc() failed", name);
+    {
+      error (0, errno, "%s: malloc() failed", name);
+      return NULL;
+    }
   buff[0] = NULL;
   struct dirent *ent = NULL;
   DIR *dir = opendir (name);
@@ -305,10 +315,10 @@ list_dir (char *name, int sort)
     {
       char *dname = ent->d_name;
       char *fname;		// full name
-      if (n == UINT_MAX)
+      if (n == INT_MAX)
 	{
-	  error (0, 0, "%s: more than %i files", name, INT_MAX - 1);
-	  return NULL;
+	  error (0, 0, "%s: more than %u files", name, INT_MAX - 1);
+	  break;
 	}
       /* ignore "." and ".." */
       if (0 == strcmp (dname, ".") || 0 == strcmp (dname, ".."))
@@ -322,10 +332,8 @@ list_dir (char *name, int sort)
 	  char **nbuff = realloc (buff, (n + 2 + BUFFSTEP) * sizeof (*buff));
 	  if (!nbuff)
 	    {
-	      /* TODO : don't abort, because it's maybe just that this dir
-	       * is too large */
-	      error (1, errno, "%s: realloc() failed", name);
-	      return NULL;
+	      error (0, errno, "%s: realloc() failed", name);
+	      break;
 	    }
 	  buff = nbuff;
 	}
@@ -333,14 +341,17 @@ list_dir (char *name, int sort)
       fname = malloc (namel + strlen (dname) + 2);
       if (!fname)
 	{
-	  error (1, errno, "%s: malloc() failed", name);
-	  return NULL;
+	  error (0, errno, "%s: malloc() failed", name);
+	  /* Try with the next file */
+	  n--;
+	  continue;
 	}
       strcpy (fname, name);
       fname[namel] = '/';
       strcpy (fname + namel + 1, dname);
       buff[n] = fname;
     }
+  /* They are some breaks in the for loop */
   if (sort)
     qsort (buff, n, sizeof (*buff), &atimesort);
   /* A NULL entry will mark the end of the buffer */
